/* This file is part of Legendes II */
/* Copyright (c) 1998 by
 *   Philippe Clavel (philippe.clavel@supelec.fr)
 *   Benjamin Lerman (benjamin.lerman@ens.fr)
 *   David Madore (david.madore@ens.fr) */

/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* File server.c */

/* Thus speaketh ye mighty CVS:
 * $Header: /home/david/cvsroot/leglab/server.c,v 1.24 1999/05/01 15:01:18 david Exp $ */

/* This program opens a network socket (on port PORTNUM) and listens
 * to it.  That's about all.  It's still pretty complicated because
 * it's threaded.  We have a first bunch of threads, called listener
 * threads (LISTENER_THREADS_NB of them), which take care of reading
 * whatever arrives on the socket and copying it to data buffers
 * associated to each connection.  A second bunch of threads, handler
 * threads (HANDLER_THREADS_NB of them) reads these buffers and
 * interprets them.  The complication comes from the fact that within
 * each of these groups the threads are undifferentiated: there isn't
 * one listener thread per connection or some such thing, but rather,
 * each thread takes care of all connections at once - so we have to
 * set up a lot of mutexes and complicated rules to stop them from
 * stepping on each others' toes. */

/* CVS identity of this file. */
static char id[] = "$Id: server.c,v 1.24 1999/05/01 15:01:18 david Exp $";

/* Include files */

#ifndef _THREAD_SAFE
#define _THREAD_SAFE		/* Make sure libc is reentrant. */
#endif

#include <stdio.h>		/* Standard I/O */
#include <stdlib.h>		/* Standard library functions */
#include <stdint.h>		/* ISO C9x types such as uint32_t */

#include <errno.h>		/* Error numbers */
#include <string.h>		/* String functions such as strerror() */

#include <sys/types.h>		/* System types */
#include <sys/socket.h>		/* Socket types */
#include <sys/time.h>		/* For struct timeval */
#include <unistd.h>		/* UNIX standard functions */
#include <fcntl.h>		/* For fcntl() and its constants */

#include <netinet/in.h>		/* INET domain (TCP/IP) sockets */

#include <netdb.h>		/* gethostbyaddr() and friends */

#include <pthread.h>		/* Posix threads */

/* Debug level */

#ifndef DEBUG_LEVEL
#define DEBUG_LEVEL 0
#endif

/* Number of threads */

#ifndef LISTENER_THREADS_NB
#define LISTENER_THREADS_NB 10	/* Number of listener threads */
#endif

#ifndef HANDLER_THREADS_NB
#define HANDLER_THREADS_NB 10	/* Number of handler threads */
#endif

/* Socket(s) and all */

#ifndef PORTNUM
#define PORTNUM 8888		/* The TCP port we listen on */
#endif

/* The ``global'' socket is the socket which we create with socket()
 * and where we listen() for new connections.  Each new accept()ed
 * connection gives birth to a new, ``local'' socket.  See the
 * accept(2) man page for details. */
int gsock;			/* The global socket descriptor */

#ifndef BUFSIZE
#define BUFSIZE 1024		/* The connection buffer size */
#endif

/* lsock_tab is the table of connections and lsock_nb is the number of
 * connections (i.e. the length of lsock_tab).  Each entry in the
 * lsock_tab is a *pointer* to a structure (a struct lsock_s) which
 * describes the connection.  Each lsock_s structure contains a local
 * socket descriptor, a data buffer, a few other associated variables,
 * and very importantly a mutex protecting all these data (see
 * NOTE0001 below for an explanation of how these mutexes work). */
int lsock_nb = 0;		/* Number of connections */
struct lsock_s {
  pthread_mutex_t d_mutex;	/* Mutex protecting all the data below */
  char active;			/* Whether connection is active */
  int s;			/* Local socket descriptor */
  struct sockaddr_in remote;	/* Socket on remote host */
  char newdata;			/* New data has arrived since a
                                 * handler thread last examined buffer */
  int buflen;			/* In bytes */
  char buffer[BUFSIZE];		/* Data buffer */
} **lsock_tab = NULL; /* Table of local sockets; note ** and not just *... */

pthread_mutex_t lsock_mutex = PTHREAD_MUTEX_INITIALIZER;

/* [NOTE0001] The lsock_mutex is the most important mutex in this
 * program.  It protects the lsock_tab.  In other words, a thread is
 * not allowed to modify any entry of the lsock_tab nor its size
 * (i.e. lsock_nb) unless it is holding the lsock_mutex.  For this
 * reason also, many loops on the connections are done with the
 * lsock_mutex held (so the lsock_tab won't be modified under the
 * looper's feet).  This means that the lsock_mutex is a pretty
 * precious resource (and perhaps we should consider replacing it by a
 * rwlock - some profiling obviously needs to be done there).
 * Besides, some other data are protected by the lsock_mutex, such as
 * the connection count, below.  However, it isn't *that* precious in
 * the sense that once a thread has found the pointer (typically
 * called sck in the code) to a connection structure, it need no
 * longer hold the lsock_mutex in order to access the data in that
 * structure.  Indeed, these data are protected by another mutex,
 * namely the d_mutex of the structure.  This d_mutex must be held
 * whenever any data in the structure is to be modified (this includes
 * reading or writing the local socket).  IMPORTANT NOTICE: How does
 * one actually get access to data within a connection?  First, one
 * locks the lsock_mutex and reads the lsock_tab to find the sck
 * pointer to the connection structure.  Then one has to lock the
 * d_mutex.  Now there is a catch there.  (1) The d_mutex must be
 * locked *before* the lsock_mutex is released, because otherwise the
 * connection structure might be free()d and destroyed between the
 * time the lsock_mutex is released and the time the d_mutex is held
 * (for the moment, we don't free() connection structures but that
 * will come, and the only guarantee we make is that the free()ing
 * thread will hold both the lsock_mutex and the d_mutex - so a thread
 * wanting to keep a connection structure must always be holding to
 * one of these mutexes).  But (2) it is FORBIDDEN to
 * pthread_mutex_lock() the d_mutex while the lsock_mutex is held
 * because that may block and the lsock_mutex is too precious to be
 * held by a blocking thread.  So the thread must
 * pthread_mutex_trylock() instead. */

/* Connection count.  The following variable is incremented every time
 * a connection is received or closed or free()d (the latter can't
 * happen yet).  To change conn_count you must first acquire the
 * lsock_mutex. */
long conn_count = 0;

/* [NOTE0008] The following three variables are used to wake up the
 * handler threads.  Indeed, the handler threads can't spend their
 * time looking for the presence of new data because that would be
 * busy-waiting.  Instead, they sleep and they are awoken when new
 * data has arrived.  The subtle point is to know when they can be put
 * to sleep again: essentially, that would be when the newdata entries
 * of all the connection structures are equal to 0.  But how do we
 * know that this is the case (we can't check them one after the other
 * because they may be coming back to 1 after we checked - and we
 * can't afford to hold the lsock_mutex *plus* every single d_mutex to
 * be sure that nothing can move)?  Well, we rule that every time a
 * listener thread finds something new (a new connection, a closed
 * connection, or data read on some connection) it must increment the
 * handler_wake_count variable and signal the handler_wake_cond
 * condition (and it must do so *after* it has treated the event).
 * The handler_wake_mutex protects both the handler_wake_cond and the
 * handler_wake_count: so a listener thread will first hold the mutex,
 * then increment the count, then signal the condition, then unlock
 * the mutex.  (No other mutex should be held during the process, and
 * especially not the lsock_mutex, because otherwise we might block
 * and possibly deadlock - and we also can't trylock() because this is
 * not something we can give up on.)  A listener thread will memorise
 * the handler_wake_count at the beginning of its job, and when it
 * thinks it has done all it should do, it will hold the mutex,
 * compare the count with what it had memorized: if the count has
 * changed, it unlocks the mutex and does the necessary job; if the
 * count hasn't changed, it will *atomically* unlock the mutex and
 * begin sleeping (as per pthread_cond_wait()). */

pthread_cond_t handler_wake_cond = PTHREAD_COND_INITIALIZER;
pthread_mutex_t handler_wake_mutex = PTHREAD_MUTEX_INITIALIZER;
long handler_wake_count = 0;

/* Useful functions */

/* ``Safe'' pthread functions.  I write those because I can't affort
 * to write the error checking clause *every* damn time I lock or
 * unlock a mutex. */

inline void
safe_pthread_mutex_init(pthread_mutex_t *mutex,
			const pthread_mutexattr_t *mutexattr)
{
  int error;
  if ((error=pthread_mutex_init(mutex,mutexattr)))
    {
      fprintf(stderr,"Initializing mutex: %s\n",strerror(error));
      exit(1);
    }
}

inline void
safe_pthread_mutex_lock(pthread_mutex_t *mutex)
{
  int error;
  if ((error=pthread_mutex_lock(mutex)))
    {
      fprintf(stderr,"Locking mutex: %s\n",strerror(error));
      exit(1);
    }
}

inline int
safe_pthread_mutex_trylock(pthread_mutex_t *mutex)
{
  int error;
  if ((error=pthread_mutex_trylock(mutex)))
    {
      if (error==EBUSY) return EBUSY;
      fprintf(stderr,"Locking mutex: %s\n",strerror(error));
      exit(1);
    }
  return 0;
}

inline void
safe_pthread_mutex_unlock(pthread_mutex_t *mutex)
{
  int error;
  if ((error=pthread_mutex_unlock(mutex)))
    {
      fprintf(stderr,"Unlocking mutex: %s\n",strerror(error));
      exit(1);
    }
}

/* Main functions */

void
init_gsock(void)
     /* Create the initial TCP/IP socket, bind() it to port PORTNUM,
      * and start listen()ing to it.  This function is exectuted in
      * single thread mode. */
{
  struct sockaddr_in name; /* To store socket ``name'' in */

  /* Create socket. */
  if ((gsock=socket(AF_INET,SOCK_STREAM,0))<0)
    {
      perror("Can't create socket()");
      exit(1);
    }
  /* Make it non-blocking.  See NOTE0006 for why we need this. */
  if ( fcntl (gsock, F_SETFL, O_NONBLOCK) < 0 )
    {
      perror("Can't make socket non-blocking");
      exit(1);
    }
  /* Bind it to all interfaces, TCP port PORTNUM. */
  memset(&name,0,sizeof(name));
  name.sin_family = AF_INET;
  name.sin_port = htons(PORTNUM); /* TCP port number for our server */
  {
    uint32_t n = INADDR_ANY;
    memcpy(&name.sin_addr,&n,sizeof(uint32_t)); /* Accept all interfaces */
  }
  if (bind(gsock,(struct sockaddr *)&name,sizeof(name))<0)
    {
      perror("Can't bind() socket");
      exit(1);
    }
  /* Start listen()ing. */
  if (listen(gsock,10)<0)
    {
      perror("Can't listen() to socket");
      exit(1);
    }
}

void
listener_loop(void)
     /* This function performs a loop that accepts incoming
      * connections and receives data from already established
      * connections. */
{
  fd_set fs;
  int s;
  struct sockaddr_in other;
  int other_namelen;
#if (DEBUG_LEVEL>=3)
  struct hostent *other_host;
#endif /* DEBUG_LEVEL>=3 */
  int max;
  int len;
  int i;
  struct timeval tv;
  struct lsock_s *sck;
  long prior_conn_count;
  char suresel;

  while (1)
    {
      /* First we select() - that is, we wait for something to
       * happen. */
#if (DEBUG_LEVEL>=5)
      fprintf(stderr,"[%0#10lx]Waiting for something to happen...\n",
	      pthread_self());
#endif /* DEBUG_LEVEL>=5 */
      /* Describe what we want to wait for.  We put every local
       * socket's desciptor in the fd_set for select() (as for max, it
       * contains the highest value of a file descriptor in the set).
       * Of course, we must hold the lsock_mutex to do all this. */
      max = gsock;
      FD_ZERO(&fs);
      FD_SET(gsock,&fs);
      safe_pthread_mutex_lock(&lsock_mutex);
      prior_conn_count = conn_count;
      for (i=0;i<lsock_nb;i++)
	{
	  if (lsock_tab[i]->active)
	    {
	      if (lsock_tab[i]->s > max)
		max = lsock_tab[i]->s;
	      FD_SET(lsock_tab[i]->s,&fs);
	    }
	}
      safe_pthread_mutex_unlock(&lsock_mutex);
      /* [NOTE0002] We have now released the lsock_mutex (of course we
       * can't afford to block on select() while holding such a
       * precious resource!).  This means that the lsock_tab may have
       * changed a lot when we emerge from our sleep.  It may even
       * change between the time when we release the lsock_mutex and
       * the time we start the select() (i.e. while you are reading
       * this :-).  It is even possible for some file desciptors from
       * the table to be closed, or even reused in that time.  This is
       * not a serious problem (we just use select() as an indication
       * of what we should do, not as an authoritative information),
       * but we should be aware of this fact. */
      /* Now do the waiting. */
      tv.tv_sec = 2; /* Timeout */
      tv.tv_usec = 0;
      if ( ( i = select (max+1, &fs, NULL, NULL, &tv) ) < 0 )
	{
	  if ( errno == EBADF )
	    /* This means that some file descriptor was no longer
	     * valid by the time we reached the select() (see
	     * NOTE0002).  This is not serious.  We just ignore and
	     * redo the select(). */
	    {
#if (DEBUG_LEVEL>=4)
	      fprintf(stderr,"[%0#10lx]Somebody stole my file descriptor :-(\n",
		      pthread_self());
#endif /* DEBUG_LEVEL>=4 */
	      sched_yield();
	      continue;
	    }
	  perror("select()ing");
	  exit(1);
	}
      if (i==0)
	/* Nothing happened (i.e. select() timed out.  We start
         * again. */
	{
	  sched_yield();
	  continue;
	}
      /* Examine what we have found. */
      /* [NOTE0004] While it is true that select() will never wake up
       * more than one sleeping thread for the same file descriptor,
       * yet we can't use that fact, because our threads are not
       * necessarily all sleeping at the same time.  So thread A may
       * be awoken for file descriptor f by select(), and before it
       * has time to take care of f, thread B may start select()ing
       * and be also awoken for the same descriptor.  So one of the
       * two is going to find out, by the time it gets to the file
       * descriptor, that the latter has nothing to say after all,
       * despite select() having indicated it as ready.  So even if
       * the file descriptor is set in fs, we must check that there
       * really is something to be found there.  An even more devilish
       * problem, see NOTE0002, is that a file descriptor may have the
       * same number as the one we selected on but actually refer to
       * something completely different (the former one has been
       * closed and the OS has reused the descriptor number).  So,
       * caveat auditor. */
      if (FD_ISSET(gsock,&fs))
	/* Global socket has something to say => new connection
         * possibly received */
	{
#if (DEBUG_LEVEL>=3)
	  fprintf(stderr,"[%0#10lx]Accepting new connection.\n",
		  pthread_self());
#endif /* DEBUG_LEVEL>=3 */
	  other_namelen = sizeof(other);
	  if ( ( s = accept (gsock, &other, &other_namelen) ) < 0 )
	    {
	      if ( errno == EAGAIN )
		/* [NOTE0006] As indicated in NOTE0004, we don't
		 * necessarily have a connection.  Here, someone stole
		 * it from us.  This is why we had to make gsock
		 * non-blocking (otherwise, the thread would have
		 * started sleeping and it would have essentially
		 * died). */
		{
#if (DEBUG_LEVEL>=4)
		  fprintf(stderr,"[%0#10lx]Somebody stole my connection :-(\n",
			  pthread_self());
#endif /* DEBUG_LEVEL>=4 */
		  sched_yield();
		  continue;
		}
	      perror("Can't accept()");
	      exit(1);
	    }
#if (DEBUG_LEVEL>=3)
	  fprintf(stderr,"[%0#10lx]Received connection:\n",
		  pthread_self());
	  fprintf(stderr,"[%0#10lx]Port number %d\n",
		  pthread_self(),other.sin_port);
	  /* XXX - gethostbyaddr() is not reentrant.  We should use
	   * gethostbyaddr_r() instead, shouldn't we?  Ugh... */
	  if ((other_host=gethostbyaddr((const char *)&other.sin_addr,
					sizeof(other.sin_addr),
					AF_INET))==NULL)
	    {
	      herror("Can't resolve address");
	    }
	  else
	    {
	      fprintf(stderr,"[%0#10lx]Host name: %s\n",
		      pthread_self(),other_host->h_name);
	    }
#endif /* DEBUG_LEVEL>=3 */
	  /* Add this new connection to the list of connections. */
	  safe_pthread_mutex_lock(&lsock_mutex);
	  if (((lsock_tab = realloc(lsock_tab,
				    (++lsock_nb)*sizeof(struct lsock_s *)))
	       /* realloc() the lsock_tab to make space for a new
                * pointer */
	       ==NULL)||
	      ((lsock_tab[lsock_nb-1] = malloc(sizeof(struct lsock_s)))
	       /* And allocate something at the end of this new
                * pointer */
	       ==NULL))
	    {
	      perror("Out of memory");
	      exit(1);
	    }
	  /* Advertise */
	  conn_count++; /* lsock_mutex is held */
	  /* Now initialize the data entries */
	  safe_pthread_mutex_init(&lsock_tab[lsock_nb-1]->d_mutex,NULL);
	  lsock_tab[lsock_nb-1]->active = 1;
	  lsock_tab[lsock_nb-1]->s = s;
	  lsock_tab[lsock_nb-1]->buflen = 0;
	  lsock_tab[lsock_nb-1]->newdata = 1;
#if (DEBUG_LEVEL>=3)
	  fprintf(stderr,"[%0#10lx]We now have %d open connections.\n",
		  pthread_self(),lsock_nb);
#endif /* DEBUG_LEVEL>=3 */
	  /* Now release the lock.  !!! Do not do this earlier, for we
	   * use lsock_nb! */
	  safe_pthread_mutex_unlock(&lsock_mutex);
	  /* Wake up the handler threads.  Do this *after* the
           * lsock_mutex has been unlocked to avoid sleeping with the
           * lsock_mutex locked (see NOTE0008). */
	  safe_pthread_mutex_lock (&handler_wake_mutex);
	  handler_wake_count++;
	  pthread_cond_signal (&handler_wake_cond);
	  safe_pthread_mutex_unlock (&handler_wake_mutex);
	}
      else
	/* Data received on some socket */
	{
#if (DEBUG_LEVEL>=4)
	  fprintf(stderr,"[%0#10lx]Received data on some socket.\n",
		  pthread_self());
#endif /* DEBUG_LEVEL>=4 */
	  /* Find out which socket spoke */
	  sck=NULL;
	  safe_pthread_mutex_lock(&lsock_mutex);
	  if ( conn_count != prior_conn_count )
	    /* Some connection has been accepted or closed while we
	     * weren't paying attention.  I.e. we're out of date.
	     * It's better to give up and redo the select().
	     * Actually, it wouldn't really hurt to remove this whole
	     * test, but I think it's safer like that. */
	    {
	      safe_pthread_mutex_unlock(&lsock_mutex);
#if ( DEBUG_LEVEL >= 3 )
	      fprintf(stderr,"[%0#10lx]Warning: overwhelmed by novelty.\n",
		      pthread_self());
#endif /* DEBUG_LEVEL >= 3 */
	      sched_yield();
	      continue;
	    }
	  for ( i=0 ; i<lsock_nb ; i++)
	    {
	      if ((lsock_tab[i]->active)&&(FD_ISSET(lsock_tab[i]->s,&fs)))
		/* Found the socket which spoke. */
		{
		  sck = lsock_tab[i];
		  break;
		}
	    }
	  if (!sck)
	    /* None did.  This is strange.  I think this shouldn't
	     * happen (this should be caught by the conn_count test
	     * above).  If it does, it is probably a misunderstanding
	     * on my part of what select() does. */
	    {
	      safe_pthread_mutex_unlock(&lsock_mutex);
#if (DEBUG_LEVEL>=1)
	      fprintf(stderr,"[%0#10lx]Warning: select() selected nothing.\n",
		      pthread_self());
#endif /* DEBUG_LEVEL>=1 */
	      sched_yield();
	      continue;
	    }
	  if (safe_pthread_mutex_trylock(&sck->d_mutex))
	    /* ...trylock() and not ...lock(), because we're not
             * allowed to block while owning the lsock_mutex (this
             * thing is _precious_, see NOTE0001). */
	    {
	      safe_pthread_mutex_unlock(&lsock_mutex);
#if (DEBUG_LEVEL>=3)
	      fprintf(stderr,"[%0#10lx]Warning: can't acquire d_mutex, giving up.\n",
		      pthread_self());
#endif /* DEBUG_LEVEL>=3 */
	      sched_yield();
	      continue;
	    }
	  safe_pthread_mutex_unlock(&lsock_mutex);
	  /* As pointed out (see NOTE0001), we don't need to hold
	   * lsock_mutex any longer to access the data in sck, now
	   * that we have the d_mutex.  And on the other hand, we
	   * could not release this sooner because we didn't have the
	   * d_mutex yet. */
	  if (!sck->active)
	    /* [NOTE0005] Someone has already closed the connection.
	     * So we have nothing to do, and we do it. */
	    {
	      safe_pthread_mutex_unlock(&sck->d_mutex);
#if (DEBUG_LEVEL>=3)
	      fprintf(stderr,"[%0#10lx]Warning: connection already obsoleted.\n",
		      pthread_self());
#endif /* DEBUG_LEVEL>=3 */
	      sched_yield();
	      continue;
	    }
	  s = sck->s; /* The socket descriptor */
	  len = BUFSIZE - sck->buflen; /* How much we try to read */
	  if ( len == 0 )
	    /* Buffer full.  We should give the handler threads some
	     * time to digest. */
	    {
	      safe_pthread_mutex_unlock(&sck->d_mutex);
#if (DEBUG_LEVEL>=3)
	      fprintf(stderr,"[%0#10lx]Warning: write buffer full.\n",
		      pthread_self());
#endif /* DEBUG_LEVEL>=3 */
	      sched_yield();
	      continue;
	    }
	  suresel = 0;
	suresel_label:
	  if ( ( i = read (s, &(sck->buffer[sck->buflen]), len) ) < 0 )
	    {
	      perror("read()ing from socket");
	      exit(1);
	    }
	  if ( i==0 )
	    /* [NOTE0003] Memento: when a socket is shut down by the
	     * other party, select() always finds something
	     * interesting on that socket, but a read() returns 0.
	     * This is the only way to know a socket has shut down,
	     * apparently (.vomit.).  So getting 0 as a result of
	     * read() is suspicious.  But we can only be sure that the
	     * socket was actually closed if the select() has been
	     * performed with the d_mutex held, because otherwise some
	     * other thread may have simply snatched the data from us.
	     * The suresel variable indicates whether we have
	     * performed a select() with the d_mutex held.  So when we
	     * read 0 (i.e. now), either suresel is false in which
	     * case we make it true (by doing the select() in
	     * question) and retry, or it is already true and then we
	     * can safely close the socket.  Disgusting?  You bet.
	     * Blame it on the Unix / pthreads interface (``seamless''
	     * is not precisely the word which comes to my mind). */
	    {
	      if ( suresel )
		/* No doubt about it, the connection was closed. */
		{
#if (DEBUG_LEVEL>=3)
		  fprintf (stderr, "[%0#10lx]Closing a connection.\n",
			   pthread_self ());
#endif /* DEBUG_LEVEL>=3 */
		  sck->active = 0;
		  shutdown (s, 2);
		  close (s);
		  safe_pthread_mutex_unlock (&sck->d_mutex);
		  /* Of course, release the d_mutex before locking the
		   * lsock_mutex! */
		  safe_pthread_mutex_lock (&lsock_mutex);
		  conn_count++; /* While holding the lsock_mutex */
		  safe_pthread_mutex_unlock (&lsock_mutex);
		  /* Wake up the handler threads, see NOTE0008. */
		  safe_pthread_mutex_lock (&handler_wake_mutex);
		  handler_wake_count++;
		  pthread_cond_signal (&handler_wake_cond);
		  safe_pthread_mutex_unlock (&handler_wake_mutex);
		  sched_yield ();
		  continue;
		}
	      else
		/* We aren't *certain* that the connection has closed.
		 * We must perform another select() operation, with
		 * zero timeout, and *while* holding the d_mutex. */
		{
		  FD_ZERO (&fs);
		  FD_SET (s,&fs);
		  tv.tv_sec = 0; /* Zero timeout */
		  tv.tv_usec = 0;
		  if ( ( i = select (max+1, &fs, NULL, NULL, &tv) ) < 0 )
		    {
		      perror ("select()ing");
		      exit(1);
		    }
		  if ( i == 1 )
		    /* Now we are sure that select() selects the socket.
		     * So we can retry the read(). */
		    {
		      suresel = 1;
		      goto suresel_label;
		      /* Yucky?  Yup.  Totally wirth-less.  ;-) */
		    }
		  else
		    /* False alarm. */
		    {
		      safe_pthread_mutex_unlock(&sck->d_mutex);
#if (DEBUG_LEVEL>=4)
		      fprintf (stderr, "[%0#10lx]Data snatched away from me.\n",
			       pthread_self ());
#endif /* DEBUG_LEVEL>=4 */
		    }
		  sched_yield();
		  continue;
		}
	    }
	  /* At this point, we have really read data from the
           * socket. */
#if (DEBUG_LEVEL>=4)
	  fprintf(stderr,"[%0#10lx]Read %d bytes from socket.\n",
		  pthread_self(),i);
#endif /* DEBUG_LEVEL>=4 */
	  sck->buflen += i;
	  sck->newdata = 1;
	  safe_pthread_mutex_unlock(&sck->d_mutex);
	  /* Wake up the handler threads.  Do this *after* the d_mutex
           * has been unlocked to avoid sleeping with the d_mutex
           * locked, see NOTE0008. */
	  safe_pthread_mutex_lock (&handler_wake_mutex);
	  handler_wake_count++;
	  pthread_cond_signal (&handler_wake_cond);
	  safe_pthread_mutex_unlock (&handler_wake_mutex);
	}
    }
}

void
handler_loop(void)
     /* This function performs a loop that handles the requests that
      * the listener threads have put in the connection buffers.  See
      * NOTE0008 for an overview of how this function works (wrt
      * sleeping, that is). */
{
  int i;
  struct lsock_s *sck;
  char some_newdata;
  long prior_count;

  safe_pthread_mutex_lock (&handler_wake_mutex);
  i = 0;
  some_newdata = 0;
  prior_count = handler_wake_count;
  safe_pthread_mutex_unlock (&handler_wake_mutex);
  while (1)
    {
      /* The trick here is that we don't want to hold the lsock_mutex
       * for too long as it is a precious resource.  So we do a loop
       * that keeps locking it and unlocking it.  I wonder if this is
       * really better, thought.  Maybe a rwlock would be more
       * appropriate than a mutex (see NOTE0001). */
      safe_pthread_mutex_lock (&lsock_mutex);
      if ( i >= lsock_nb )
	/* Looped around all connections (actually, we may have missed
	 * some since we weren't holdong the lsock_mutex all the time,
	 * but that doesn't really matter, we'll get them next
	 * time). */
	{
	  safe_pthread_mutex_unlock (&lsock_mutex);
	  if ( !some_newdata )
	    /* Hmmm...  We looped around all connections and found
	     * nothing new at all.  Perhaps we finished our task.  We
	     * need to make sure, though, because we may have missed
	     * some connections.  So we check that nothing has
	     * happened behind our back during this loop. */
	    {
	      safe_pthread_mutex_lock (&handler_wake_mutex);
	      if ( handler_wake_count == prior_count )
		/* Indeed, no new data has arrived during the entire
		 * loop.  So we are quite finished, and we can go back
		 * to sleep.  We will release the handler_wake_mutex
		 * atomically as we start sleeping. */
		goto handler_sleep;
	      /* Keep holding on that mutex, baby. */
	    }
	  else
	    safe_pthread_mutex_lock (&handler_wake_mutex);
	  i = 0;
	  some_newdata = 0;
	  prior_count = handler_wake_count;
	  safe_pthread_mutex_unlock (&handler_wake_mutex);
	  sched_yield();
	  continue;
	}
      sck = lsock_tab[i];
      if ( sck->newdata )
	/* [NOTE0007] We perform this test without holding the d_mutex
	 * (still we hold the lsock_mutex otherwise the sck
	 * dereference would be forbidden, see NOTE0001).  This is
	 * because we can't guarantee that we will be able to hold the
	 * d_mutex at all.  Still, we need to have some idea of
	 * whether there actually is something new here.  Perhaps we
	 * will read 1 when in fact it was set to 0: that is
	 * unimportant.  On the other hand, when sck->newdata is set
	 * to 1 by a listener thread, the latter also increases the
	 * handler_wake_count, so if we miss that we will notice
	 * something when we check handler_wake_count.  To summarize,
	 * any value of sck->newdata later than the last time we
	 * unlocked the handler_wake_mutex will do, and we are
	 * guaranteed to get such a value.  (Moreover, we can't have
	 * any byte tearing because sck->newdata is a char.)  So this
	 * is OK.  (Would somebody please confirm...) */
	some_newdata = 1;
      if ( safe_pthread_mutex_trylock (&sck->d_mutex) )
	{
	  /* Someone is doing something with this connection already.
	   * We go on. */
	  safe_pthread_mutex_unlock (&lsock_mutex);
	  i++;
	  sched_yield();
	  continue;
	}
      /* Now we have the d_mutex so we can release the lsock_mutex to
       * access sck (see NOTE0001). */
      safe_pthread_mutex_unlock (&lsock_mutex);
      /* This is a stupid reading function: we just gobble all the
       * data and throw it to stdout.  Someday we will have something
       * smarter here. */
      if ( sck->buflen > 0 )
	{
#if ( DEBUG_LEVEL >= 1 )
	  fprintf (stderr, "[%0#10lx]Gobbling %d bytes of data.\n",
		   pthread_self(), sck->buflen);
#endif /* DEBUG_LEVEL >= 1 */
	  fwrite (sck->buffer, sizeof(char), sck->buflen, stdout);
	  fflush (stdout);
	  sck->buflen = 0;
	}
      sck->newdata = 0;
      /* Setting sck->newdata to 0 had better be the last thing we do
       * before we release the d_mutex because when we test for
       * sck->newdata at point NOTE0007 above we aren't holding to the
       * d_mutex.  This is not critical but it is probably better. */
      safe_pthread_mutex_unlock (&sck->d_mutex);
      i++;
      continue;
    handler_sleep:
      /* When we arrive here we *must* be owning the
       * handler_wake_mutex. */
      pthread_cond_wait (&handler_wake_cond, &handler_wake_mutex);
      safe_pthread_mutex_unlock (&handler_wake_mutex);
    }
}

void *
entry_listener(void *garbage)
     /* Entry point for a listener thread */
{
  listener_loop();
  return NULL;
}

void *
entry_handler(void *garbage)
     /* Entry point for a handler thread */
{
  handler_loop();
  return NULL;
}

int
main(void)
     /* Abandon all hope, ye who enter here */
{
  pthread_t garbage;
  pthread_attr_t attr;
  int i;
  int error;

  /* Initialize global socket */
  init_gsock();
  /* Create listener threads */
  for (i=0;i<LISTENER_THREADS_NB;i++)
    {
      pthread_attr_init(&attr);
      /* We create our threads in the detached state because we won't
       * try to join them again (see pthread_join(3) for an
       * explanation. */
      pthread_attr_setdetachstate(&attr,PTHREAD_CREATE_DETACHED);
      if ((error=pthread_create(&garbage,&attr,entry_listener,NULL)))
	{
	  fprintf(stderr,"Creating listener thread: %s\n",strerror(error));
	  exit(1);
	}
      pthread_attr_destroy(&attr);
    }
  /* Create handler threads */
  for (i=0;i<HANDLER_THREADS_NB;i++)
    {
      pthread_attr_init(&attr);
      /* We create our threads in the detached state because we won't
       * try to join them again (see pthread_join(3) for an
       * explanation. */
      pthread_attr_setdetachstate(&attr,PTHREAD_CREATE_DETACHED);
      if ((error=pthread_create(&garbage,&attr,entry_handler,NULL)))
	{
	  fprintf(stderr,"Creating handler thread: %s\n",strerror(error));
	  exit(1);
	}
      pthread_attr_destroy(&attr);
    }
  /* The father thread dies here */
  pthread_exit(NULL);
  return 0; /* Keep gcc happy */
}
